﻿using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.EvolutionTracking {
  [StorableClass]
  [Item("GenealogyAnalyzer", "An analyzer which performs the necessary instrumentation to record the evolution of a genetic algorithm.")]
  public class GenealogyAnalyzer<T> : SingleSuccessorOperator, IAnalyzer
  where T : class,IItem {

    public string CrossoverParentsParameterName { get; set; }
    public string CrossoverChildParameterName { get; set; }
    public string ManipulatorChildParameterName { get; set; }

    public IScopeTreeLookupParameter<DoubleValue> QualityParameter;
    public IScopeTreeLookupParameter<T> PopulationParameter;

    public ICrossoverOperator<T> BeforeCrossoverOperator { get; set; }
    public ICrossoverOperator<T> AfterCrossoverOperator { get; set; }
    public IManipulatorOperator<T> BeforeManipulatorOperator { get; set; }
    public IManipulatorOperator<T> AfterManipulatorOperator { get; set; }

    private const string PopulationParameterName = "Population";
    private const string QualityParameterName = "Quality";

    private const string GenerationsParameterName = "Generations";
    private const string ResultsParameterName = "Results";
    private const string PopulationGraphParameterName = "PopulationGraph";

    private const string CrossoverParameterName = "Crossover";
    private const string ManipulatorParameterName = "Mutator";
    private const string SolutionCreatorParameterName = "SolutionCreator";

    private const string EnableCrossoverTrackingParameterName = "EnableCrossoverTracking";
    private const string EnableManipulatorTrackingParameterName = "EnableManipulatorTracking";
    private const string EnableSolutionCreatorTrackingParameterName = "EnableSolutionCreatorTracking"; // should always be enabled. maybe superfluous

    #region parameter properties
    public ILookupParameter<ResultCollection> ResultsParameter {
      get { return (ILookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    public ILookupParameter<IntValue> GenerationsParameter {
      get { return (ILookupParameter<IntValue>)Parameters[GenerationsParameterName]; }
    }
    public IValueParameter<BoolValue> EnableCrossoverTrackingParameter {
      get { return (IValueParameter<BoolValue>)Parameters[EnableCrossoverTrackingParameterName]; }
    }
    public IValueParameter<BoolValue> EnableManipulatorTrackingParameter {
      get { return (IValueParameter<BoolValue>)Parameters[EnableManipulatorTrackingParameterName]; }
    }
    public IValueParameter<BoolValue> EnableSolutionCreatorTrackingParameter {
      get { return (IValueParameter<BoolValue>)Parameters[EnableSolutionCreatorTrackingParameterName]; }
    }
    public ILookupParameter<ICrossover> CrossoverParameter {
      get { return (ILookupParameter<ICrossover>)Parameters[CrossoverParameterName]; }
    }
    public ILookupParameter<IManipulator> ManipulatorParameter {
      get { return (ILookupParameter<IManipulator>)Parameters[ManipulatorParameterName]; }
    }
    public ILookupParameter<ISolutionCreator> SolutionCreatorParameter {
      get { return (ILookupParameter<ISolutionCreator>)Parameters[SolutionCreatorParameterName]; }
    }
    #endregion

    #region properties
    public ICrossover Crossover {
      get { return CrossoverParameter.ActualValue; }
    }
    public IManipulator Manipulator {
      get { return ManipulatorParameter.ActualValue; }
    }
    public ISolutionCreator SolutionCreator {
      get { return SolutionCreatorParameter.ActualValue; }
    }
    public BoolValue EnableCrossoverTracking {
      get { return EnableCrossoverTrackingParameter.Value; }
    }
    public BoolValue EnableManipulatorTracking {
      get { return EnableManipulatorTrackingParameter.Value; }
    }
    public BoolValue EnableSolutionCreatorTracking {
      get { return EnableSolutionCreatorTrackingParameter.Value; }
    }
    public ResultCollection Results {
      get { return ResultsParameter.ActualValue; }
    }
    public IntValue Generations {
      get { return GenerationsParameter.ActualValue; }
    }
    public IGenealogyGraph GenealogyGraph {
      get {
        IResult result;
        if (!Results.ContainsKey(PopulationGraphParameterName)) {
          result = new Result(PopulationGraphParameterName, new GenealogyGraph());
          Results.Add(result);
        } else {
          result = Results[PopulationGraphParameterName];
        }
        var graph = (IGenealogyGraph)result.Value;
        return graph;
      }
    }
    #endregion

    public GenealogyAnalyzer() {
      // the instrumented operators
      Parameters.Add(new LookupParameter<ICrossover>(CrossoverParameterName, "The crossover operator."));
      Parameters.Add(new LookupParameter<IManipulator>(ManipulatorParameterName, "The manipulator operator."));
      Parameters.Add(new LookupParameter<ISolutionCreator>(SolutionCreatorParameterName, "The solution creator operator."));
      // the analyzer parameters
      Parameters.Add(new ValueParameter<BoolValue>(EnableCrossoverTrackingParameterName, new BoolValue(true)));
      Parameters.Add(new ValueParameter<BoolValue>(EnableManipulatorTrackingParameterName, new BoolValue(true)));
      Parameters.Add(new ValueParameter<BoolValue>(EnableSolutionCreatorTrackingParameterName, new BoolValue(true)));
      // parameters required by the analyzer to do its work
      Parameters.Add(new LookupParameter<IntValue>(GenerationsParameterName, "The number of generations so far."));
      Parameters.Add(new LookupParameter<ResultCollection>(ResultsParameterName));
      PopulationParameter = new ScopeTreeLookupParameter<T>(PopulationParameterName);
      QualityParameter = new ScopeTreeLookupParameter<DoubleValue>(QualityParameterName);
      Parameters.Add(PopulationParameter);
      Parameters.Add(QualityParameter);
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new GenealogyAnalyzer<T>(this, cloner);
    }
    protected GenealogyAnalyzer(GenealogyAnalyzer<T> original, Cloner cloner)
      : base(original, cloner) {
    }
    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      // the instrumented operators
      if (!Parameters.ContainsKey(CrossoverParameterName))
        Parameters.Add(new LookupParameter<ICrossover>(CrossoverParameterName, "The crossover operator."));
      if (!Parameters.ContainsKey(ManipulatorParameterName))
        Parameters.Add(new LookupParameter<IManipulator>(ManipulatorParameterName, "The manipulator operator."));
      if (!Parameters.ContainsKey(SolutionCreatorParameterName))
        Parameters.Add(new LookupParameter<ISolutionCreator>(SolutionCreatorParameterName, "The solution creator operator."));
      // the analyzer parameters
      if (!Parameters.ContainsKey(EnableCrossoverTrackingParameterName))
        Parameters.Add(new ValueParameter<BoolValue>(EnableCrossoverTrackingParameterName, new BoolValue(true)));
      if (!Parameters.ContainsKey(EnableManipulatorTrackingParameterName))
        Parameters.Add(new ValueParameter<BoolValue>(EnableManipulatorTrackingParameterName, new BoolValue(true)));
      if (!Parameters.ContainsKey(EnableSolutionCreatorTrackingParameterName))
        Parameters.Add(new ValueParameter<BoolValue>(EnableSolutionCreatorTrackingParameterName, new BoolValue(true)));
      // parameters required by the analyzer to do its work
      if (!Parameters.ContainsKey(GenerationsParameterName))
        Parameters.Add(new LookupParameter<IntValue>(GenerationsParameterName, "The number of generations so far."));
      if (!Parameters.ContainsKey(ResultsParameterName))
        Parameters.Add(new LookupParameter<ResultCollection>(ResultsParameterName));
      if (!Parameters.ContainsKey(PopulationParameterName)) {
        PopulationParameter = new ScopeTreeLookupParameter<T>(PopulationParameterName);
        Parameters.Add(PopulationParameter);
      }
      if (!Parameters.ContainsKey(QualityParameterName)) {
        QualityParameter = new ScopeTreeLookupParameter<DoubleValue>(QualityParameterName);
        Parameters.Add(QualityParameter);
      }
    }

    public bool EnabledByDefault {
      get { return false; }
    }

    private void ConfigureTrackingOperators() {
      // at the beginning we add the before/after operators to the instrumented operators
      if (Crossover != null) {
        var instrumentedCrossover = (InstrumentedOperator)Crossover;
        instrumentedCrossover.AfterExecutionOperators.Clear();
        instrumentedCrossover.BeforeExecutionOperators.Clear();

        if (EnableCrossoverTracking.Value) {
          if (BeforeCrossoverOperator != null) {
            BeforeCrossoverOperator.ParentsParameter.ActualName = CrossoverParentsParameterName;
            BeforeCrossoverOperator.ChildParameter.ActualName = CrossoverChildParameterName;
            instrumentedCrossover.BeforeExecutionOperators.Add(BeforeCrossoverOperator);
          }
          if (AfterCrossoverOperator != null) {
            AfterCrossoverOperator.ParentsParameter.ActualName = CrossoverParentsParameterName;
            AfterCrossoverOperator.ChildParameter.ActualName = CrossoverChildParameterName;
            instrumentedCrossover.AfterExecutionOperators.Add(AfterCrossoverOperator);
          }
        }
      }

      if (Manipulator != null) {
        var instrumentedManipulator = (InstrumentedOperator)Manipulator;
        instrumentedManipulator.AfterExecutionOperators.Clear();
        instrumentedManipulator.BeforeExecutionOperators.Clear();
        if (EnableManipulatorTracking.Value && Manipulator != null) {
          if (BeforeManipulatorOperator != null) {
            BeforeManipulatorOperator.ChildParameter.ActualName = ManipulatorChildParameterName;
            instrumentedManipulator.BeforeExecutionOperators.Add(BeforeManipulatorOperator);
          }
          if (AfterManipulatorOperator != null) {
            AfterManipulatorOperator.ChildParameter.ActualName = ManipulatorChildParameterName;
            instrumentedManipulator.AfterExecutionOperators.Add(AfterManipulatorOperator);
          }
        }
      }
    }

    public override IOperation Apply() {
      var population = PopulationParameter.ActualValue.ToList();
      var qualities = QualityParameter.ActualValue.ToList();

      if (Generations.Value == 0) {
        ConfigureTrackingOperators();

        foreach (var individual in population) {
          var vertex = new GenealogyGraphNode<T> { Content = individual, Rank = Generations.Value, };
          GenealogyGraph.AddVertex(vertex);
        }
      } else {
        var elite = population.FirstOrDefault(x => GenealogyGraph.Contains(x));
        if (elite != null) {
          var vertex = new GenealogyGraphNode<T> {
            Content = elite,
            Rank = Generations.Value,
            IsElite = true
          };
          var previousVertex = GenealogyGraph[elite].Last();
          GenealogyGraph.AddVertex(vertex);
          previousVertex.AddForwardArc(vertex);
          vertex.AddReverseArc(previousVertex);
        }
      }
      // update qualities
      for (int i = 0; i < population.Count; ++i) {
        var individual = population[i];
        foreach (var vertex in GenealogyGraph[individual].Cast<IGenealogyGraphNode<T>>()) {
          vertex.Quality = qualities[i].Value;
        }
      }

      return base.Apply();
    }
  }
}
